# Copyright (C) 2006-2008 by Sergio Pistone
#  sergio_pistone@yahoo.com.ar
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the
#  Free Software Foundation, Inc.,
#  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

require File.expand_path( File.dirname( __FILE__ ) + "/../utils/strings" )
require File.expand_path( File.dirname( __FILE__ ) + "/../utils/logging" )
require File.expand_path( File.dirname( __FILE__ ) + "/amarok" )

require "cgi"
require "thread"

module AmarokLyricsScript

	# Hack to make module methods become class methods when the module gets included
	def AmarokLyricsScript.included( including )
		if including.is_a?( Class )
			including.extend( ClassMethods ) # adds class methods
		else # including.is_a?( Module )
			including::ClassMethods.append_class_methods( self )
		end
	end

	# Methods under this module will became class methods when the module gets included
	# Note: don't use def self.<method name> but just <method name>
	module ClassMethods
		def ClassMethods.append_class_methods( mod )
			include mod::ClassMethods
		end

		def notify( message )
			Amarok.notify( "<b>[" + script_name() + "]</b> " + message.gsub( "\n", " " ) )
		end

		def popup( message )
			Amarok.popup( "<b>[" + script_name() + "]</b><br/>" + message )
		end

		def add_custom_menu_item( menu_item )
			Amarok.add_custom_menu_item( script_name(), menu_item )
		end

		def remove_custom_menu_item( menu_item )
			Amarok.remove_custom_menu_item( script_name(), menu_item )
		end
	end

	def script_name()
		return self.class.script_name()
	end

	def notify( message )
		self.class.notify( message )
	end

	def popup( message )
		self.class.popup( message )
	end

	def add_custom_menu_item( menu_item )
		self.class.add_custom_menu_item( menu_item )
	end

	def remove_custom_menu_item( menu_item )
		self.class.remove_custom_menu_item( menu_item )
	end

	def log?()
		return @log_file != nil
	end

	def log_file()
		return @log_file
	end

	def log_file=( log_file )
		@log_file = log_file
	end

	def log( message, new_lines=1 )
		Logging.log( @log_file, message, new_lines ) if ( @log_file )
	end

	def on_start()
		# default stub implementation
	end

	def on_quit()
		# default stub implementation
	end

	def on_configure()
		# default stub implementation
	end

	def on_fetch_lyrics( artist, title, album, year )

		# NOTE: we must check that the current song doesn't change while we're fetching
		# the lyrics, otherwise Amarok will associate the lyrics with the wrong song
		prev_artist = Strings.normalize!( Amarok.get_current_artist().to_s() )
		prev_title  = Strings.normalize!( Amarok.get_current_title().to_s() )

		lyrics_data = lyrics_full_search( artist, title, album, year )

		return if ( Strings.normalize!( Amarok.get_current_artist().to_s() ) != prev_artist ||
					Strings.normalize!( Amarok.get_current_title().to_s() ) != prev_title )

		if ( lyrics_data["lyrics"] )
			Amarok.show_lyrics(
				lyrics_data["artist"],
				lyrics_data["title"],
				lyrics_data["lyrics"],
				lyrics_data["site_name"],
				lyrics_data["site_host"],
				lyrics_data["url"],
				lyrics_data["add_url"]
			)
		elsif ( lyrics_data["suggestions"] )
			Amarok.show_suggestions(
				lyrics_data["artist"],
				lyrics_data["title"],
				lyrics_data["suggestions"],
				lyrics_data["add_url"]
			)
		else
			Amarok.show_not_found(
				lyrics_data["artist"],
				lyrics_data["title"],
				lyrics_data["add_url"]
			)
		end

	end

	def on_fetch_lyrics_from_url( url, artist, title, album, year )

		# NOTE: we must check that the current song doesn't change while we're fetching
		# the lyrics, otherwise Amarok will associate the lyrics with the wrong song
		prev_artist = Strings.normalize!( Amarok.get_current_artist().to_s() )
		prev_title  = Strings.normalize!( Amarok.get_current_title().to_s() )

		lyrics_data = lyrics_from_url( url, artist, title, album, year )

		return if ( Strings.normalize!( Amarok.get_current_artist().to_s() ) != prev_artist ||
					Strings.normalize!( Amarok.get_current_title().to_s() ) != prev_title )

		if ( lyrics_data["lyrics"] )
			Amarok.show_lyrics(
				lyrics_data["artist"],
				lyrics_data["title"],
				lyrics_data["lyrics"],
				lyrics_data["site_name"],
				lyrics_data["site_host"],
				lyrics_data["url"],
				lyrics_data["add_url"]
			)
		else
			Amarok.show_not_found(
				lyrics_data["artist"],
				lyrics_data["title"],
				lyrics_data["add_url"]
			)
		end

	end

	def set_configure_msg()
		@mutex.synchronize do
			# prioritize configure messages:
			if ( @messages[0] != "configure" )
				@messages.insert( 0, "configure" )
			end
			@queued_cond.signal()
		end
	end
	protected :set_configure_msg

	def set_generic_msg( command )
		@mutex.synchronize do
			# we queue only two messages to try to stay in synch with Amarok current track
			@messages[@messages.size < 2 ? @messages.size : 1 ] = command
			@queued_cond.signal()
		end
	end
	protected :set_generic_msg

	def get_msg()
		@mutex.synchronize do
			@queued_cond.wait( @mutex ) if ( @messages.empty?() )
			return @messages.shift()
		end
	end
	protected :get_msg

	def debug( message )
		system( "kdialog", "--sorry", "<qt>" + message.gsub( "\n", "<br/>" ) + "</qt>" )
	end

	def exec( threaded_tasks=false )
		@mutex = Mutex.new()
		@queued_cond  = ConditionVariable.new()
		@messages = []

		exit_code = 0

		a = Thread.new() do
			begin
				on_start()
				run_worker( threaded_tasks )
			rescue Errno::ECONNREFUSED => e
				popup( I18N.get( "amarok.lyricsscript.error.connectionrefused" ) )
				popup( e.backtrace().join( "<br/>" ) )
			rescue SocketError, TimeoutError => e
				popup( I18N.get( "amarok.lyricsscript.error.connectiondown" ) )
			rescue Exception => e # unexpected error
				$stderr << e.message() << "\n"
				$stderr << e.backtrace().join( "\n" )
				log( e.message() )
				log( e.backtrace().join( "\n" ), 0 )
				exit_code = 1
			end
			exit( exit_code )
		end

		trap( "SIGTERM" ) { $stderr << "SIGTERM caugth" } # causes to fall to the Exception block below on abnormal termination
		begin
			run_stdin_listener()
		rescue Exception => e
		end
		on_quit()
		exit( exit_code )
	end

	def run_stdin_listener()
		loop do
			message = gets().chomp()
			command = /[A-Za-z]*/.match( message ).to_s()
			case command
				when "configure"
					set_configure_msg()
				when "fetchLyrics"
					set_generic_msg( message )
				when "fetchLyricsByUrl"
					set_generic_msg( message )
				when "customMenuClicked"
					set_generic_msg( message )
			end
		end
	end
	protected :run_stdin_listener

	def run_worker( threaded_tasks )
		loop do
			message = get_msg()
			return if ( message == nil )
			command = /[A-Za-z]*/.match( message ).to_s()
			case command
				when "configure"
					on_configure()
				when "fetchLyrics"
					artist = Amarok.get_current_artist()
					title  = Amarok.get_current_title()
					if ( artist && title )
						if ( threaded_tasks )
							Thread.new() do
								on_fetch_lyrics(
									Strings.cleanup_artist( artist, title ),
									Strings.cleanup_title( title ),
									Amarok.get_current_album(),
									Amarok.get_current_year()
								)
							end
						else
							on_fetch_lyrics(
								Strings.cleanup_artist( artist, title ),
								Strings.cleanup_title( title ),
								Amarok.get_current_album(),
								Amarok.get_current_year()
							)
						end
					else
						Amarok.show_error()
					end
				when "fetchLyricsByUrl"
					args = message.split()
					if args.length < 2
						Amarok.show_error()
					else
						artist = Amarok.get_current_artist().to_s()
						title  = Amarok.get_current_title().to_s()
						if ( threaded_tasks )
							Thread.new() do
								on_fetch_lyrics_from_url(
									args[1],
									Strings.cleanup_artist( artist, title ),
									Strings.cleanup_title( title ),
									Amarok.get_current_album(),
									Amarok.get_current_year()
								)
							end
						else
							on_fetch_lyrics_from_url(
								args[1],
								Strings.cleanup_artist( artist, title ),
								Strings.cleanup_title( title ),
								Amarok.get_current_album(),
								Amarok.get_current_year()
							)
						end
					end
				when "customMenuClicked"
					urls_start = message.index( /([a-zA-Z]+:\/\/[^: ]+){1,}/ )
					return if ( urls_start == nil )
					menu_id = message.slice( "customMenuClicked:".size()+1..urls_start-2 )
					menu, item = Amarok.get_custom_menu_item( menu_id )
					if ( menu && item )
						urls = message.slice( urls_start..-1 ).split( " " )
						if ( threaded_tasks )
							Thread.new() { on_custom_menu_item_selected( menu, item, urls ) }
						else
							on_custom_menu_item_selected( menu, item, urls )
						end
					end
			end
		end
	end
	protected :run_worker

end
