#!/usr/bin/ruby

#  Copyright (C) 2007 by Sergio Pistone
#  sergio_pistone@yahoo.com.ar
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the
#  Free Software Foundation, Inc.,
#  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

require File.expand_path( File.dirname( __FILE__ ) + "/../utils/xmlhash" )
require File.expand_path( File.dirname( __FILE__ ) + "/../utils/logging" )
require File.expand_path( File.dirname( __FILE__ ) + "/../utils/kde" )
require File.expand_path( File.dirname( __FILE__ ) + "/../i18n/i18n" )
require File.expand_path( File.dirname( __FILE__ ) + "/../gui/gui" )
require File.expand_path( File.dirname( __FILE__ ) + "/../lyrics" )
require File.expand_path( File.dirname( __FILE__ ) + "/amarok" )
require File.expand_path( File.dirname( __FILE__ ) + "/amaroklyricsscript" )
require File.expand_path( File.dirname( __FILE__ ) + "/plugins" )

TOOLKIT_PRIORITY = ["qt3", "qt4", "gtk", "tk"]

$LOG_MAX_LINES = 2000
$LOG_FILENAME = "#{ENV["HOME"]}/.wikilyrics.log"

class PluginsManager

	include AmarokLyricsScript

	@@MENU_ENTRY_SEARCH_CURRENT = I18N.get( "amarok.pluginsmanager.search.current" )
	@@MENU_ENTRY_SEARCH_SELECTED = I18N.get( "amarok.pluginsmanager.search.selected" )
	@@MENU_ENTRY_CLEAR_LYRICS_CACHE = I18N.get( "amarok.pluginsmanager.clearlyricscache" )

	def PluginsManager.script_name()
		return "Wiki-Lyrics"
	end

	def PluginsManager.config_file()
		return "wikilyrics.xml"
	end

	def config_file()
		return self.class.config_file()
	end

	def debug( text )
		`kdialog --msgbox #{Strings.shell_quote( text )}`
	end

	def log_file=( filename )
		super( filename )
		Plugins.all_plugins().each() { |plugin| plugin.log_file = log_file() }
	end

	def cleanup_lyrics?()
		return @cleanup_lyrics
	end

	def cleanup_lyrics=( cleanup_lyrics )
		@cleanup_lyrics = cleanup_lyrics
		Plugins.all_plugins().each() { |plugin| plugin.cleanup_lyrics = cleanup_lyrics }
	end

	def initialize()
		self.cleanup_lyrics = true
		self.log_file = nil
	end

	def on_start()
		read_config()
		add_custom_menu_item( @@MENU_ENTRY_SEARCH_CURRENT )
		add_custom_menu_item( @@MENU_ENTRY_SEARCH_SELECTED )
		add_custom_menu_item( @@MENU_ENTRY_CLEAR_LYRICS_CACHE )
		Plugins.all_plugins().each() { |plugin| plugin.on_start() }
	end

	def on_quit()
		write_config()
		remove_custom_menu_item( @@MENU_ENTRY_SEARCH_CURRENT )
		remove_custom_menu_item( @@MENU_ENTRY_SEARCH_SELECTED )
		remove_custom_menu_item( @@MENU_ENTRY_CLEAR_LYRICS_CACHE )
		Plugins.all_plugins().each() { |plugin| plugin.on_quit() }
	end

	def on_custom_menu_item_selected( menu, item, urls )
		if ( menu == script_name() )
			if ( item == @@MENU_ENTRY_SEARCH_CURRENT )
				on_search_current_lyrics()
			elsif ( item == @@MENU_ENTRY_SEARCH_SELECTED )
				url = URI.parse( urls[0] )
				return if ( url.scheme != "file" )
				on_search_file_lyrics( URI.decode( url.path ) )
			elsif ( item == @@MENU_ENTRY_CLEAR_LYRICS_CACHE )
				if ( GUI.show_confirmation_dialog( I18N.get( "amarok.pluginsmanager.clearlyricscache.confirm" ), script_name()) )
					Amarok.query( "DELETE FROM lyrics WHERE url <> ''" )
					notify( I18N.get( "amarok.pluginsmanager.clearlyricscache.done" ) )
				end
			end
		elsif ( plugin = Plugins.plugin_by_name( menu ) )
			plugin.on_custom_menu_item_selected( menu, item, urls )
		end
	end

	def on_search_current_lyrics()
		if ( Amarok.playing?() )
			song_data = Amarok.query_song_data( Amarok.get_current_path() )
			if ( GUI.show_search_lyrics_dialog( song_data ) )
				on_fetch_lyrics( song_data["artist"], song_data["title"], song_data["album"], song_data["year"] )
			end
		else
			notify( I18N.get( "amarok.pluginsmanager.search.current.nosongplaying" ) )
		end
	end

	def on_search_file_lyrics( file )
		if ( (song_data = Amarok.query_song_data( file )).empty?() )
			notify( I18N.get( "amarok.pluginsmanager.search.noinfofound" ) )
		elsif ( GUI.show_search_lyrics_dialog( song_data ) )
			lyrics_data = lyrics_full_search( song_data["artist"], song_data["title"], song_data["album"], song_data["year"] )
			if ( lyrics_data["lyrics"] )
				GUI.show_lyrics_dialog( lyrics_data )
			else
				notify( I18N.get( "amarok.pluginsmanager.search.nolyricsfound", song_data["title"], song_data["artist"] ) )
			end
		end
	end

	def lyrics_from_url( url, artist, title, album, year )

		lyrics_data = Lyrics.normalize_lyrics_data( {}, artist, title, album, year )
		searched_wikis = []
		wiki_names = Plugins.wiki_names()
		used_plugins = Plugins.used_plugins()

		popup( I18N.get( "amarok.pluginsmanager.error.nopluginsselected" ) ) if ( used_plugins.size() == 0 )

		used_plugins.each() do |plugin|
			next if ( ! plugin.known_url?( url ) )
			lyrics_data = plugin.lyrics_from_url( url, artist, title, album, year )
			searched_wikis << plugin.plugin_name() if ( wiki_names.include?( plugin.plugin_name() ) )
			break if ( lyrics_data["lyrics"] )
		end

		wikis_process_lyrics_data( searched_wikis, artist, title, album, year, lyrics_data )

		return lyrics_data
	end

	def lyrics_full_search( artist, title, album, year )

		lyrics_data = Lyrics.normalize_lyrics_data( {}, artist, title, album, year )
		suggestions = []
		add_url = nil
		searched_wikis = []
		wiki_names = Plugins.wiki_names()
		used_plugins = Plugins.used_plugins()

		popup( I18N.get( "amarok.pluginsmanager.error.nopluginsselected" ) ) if ( used_plugins.size() == 0 )

		used_plugins.each() do |plugin|
			begin
				lyrics_data = plugin.lyrics_full_search( artist, title, album, year )
				if ( lyrics_data["suggestions"].size > 0 )
					suggestions << plugin.plugin_name()
					suggestions.concat( lyrics_data["suggestions"] )
				end
				searched_wikis << plugin.plugin_name() if ( wiki_names.include?( plugin.plugin_name() ) )
				add_url = lyrics_data["add_url"] if ( ! add_url && lyrics_data["add_url"] )
				break if ( lyrics_data["lyrics"] )
			rescue TimeoutError
				notify( I18N.get( "amarok.pluginsmanager.search.plugintimeout", plugin.plugin_name(), plugin.site_host() ) )
			end
		end

		lyrics_data["suggestions"] = suggestions
		lyrics_data["add_url"] = add_url

		wikis_process_lyrics_data( searched_wikis, artist, title, album, year, lyrics_data )

		return lyrics_data

	end

	def wikis_process_lyrics_data( searched_wikis, artist, title, album, year, lyrics_data )
		Plugins.wiki_plugins().each() do |plugin|
			plugin.wiki_process_lyrics_data(
				searched_wikis.include?( plugin.plugin_name() ),
				artist,
				title,
				album,
				year,
				lyrics_data,
				false
			)
		end
	end

	def on_configure()
		configure()
		write_config()
	end

	def configure()
		old_log = log?()
		values = {
			"cleanup_lyrics" => cleanup_lyrics?(),
			"write_log" => log?(),
			"used_plugins" => Plugins.used_names(),
			"script_name" => script_name(),
			"unused_plugins" => Plugins.all_names() - Plugins.used_names()
		}
		if ( GUI.show_manager_script_config_dialog( values ) )
			log = values["write_log"].to_s() == "true"
			new_log_file = log ? $LOG_FILENAME : nil
			Logging.setup( new_log_file, $LOG_MAX_LINES, log_file() )
			self.log_file = new_log_file
			self.cleanup_lyrics = values["cleanup_lyrics"].to_s() == "true"
			Plugins.used_names = values["used_plugins"]
		end
		Plugins.all_plugins().each() { |plugin| plugin.configure() }
	end

	def read_config()
		old_log = log?()
		values = {
			"cleanup_lyrics" => cleanup_lyrics?(),
			"write_log" => log?(),
			"used_plugins" => Plugins.used_names().join( ";" )
		}
		XMLHash.read( config_file(), values )
		log = values["write_log"].to_s() == "true"
		new_log_file = log ? $LOG_FILENAME : nil
		Logging.setup( new_log_file, $LOG_MAX_LINES, log_file() )
		self.log_file = new_log_file
		self.cleanup_lyrics = values["cleanup_lyrics"].to_s() == "true"
		Plugins.used_names = values["used_plugins"].split( ";" )
		Plugins.all_plugins().each() { |plugin| plugin.read_config( config_file() ) }
	end

	def write_config()
		values = {
			"cleanup_lyrics" => cleanup_lyrics?(),
			"write_log" => log?(),
			"used_plugins" => Plugins.used_names().join( ";" )
		}
		XMLHash.write( config_file(), values )
		Plugins.all_plugins().each() { |plugin| plugin.write_config( config_file() ) }
	end

end

if ( $0 == __FILE__ )

	if ( GUI.set_toolkit_priority( TOOLKIT_PRIORITY ) == nil )
		Amarok.popup( I18N.get( "amarok.pluginsmanager.error.notoolkit" ) )
		exit( 0 )
	end

	proxy_url, excluded_urls, reverse = KDE.get_proxy_settings()
	HTTP.set_proxy_settings( proxy_url, excluded_urls, reverse )

	PluginsManager.new().exec()

end
