# Copyright (C) 2006-2008 by Sergio Pistone
#  sergio_pistone@yahoo.com.ar
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the
#  Free Software Foundation, Inc.,
#  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

$LOAD_PATH << File.expand_path(File.dirname(__FILE__))

require "utils/logging"
require "utils/strings"
require "utils/http"

$LONG_LOG = false
$DFLT_LOG = false
$DFLT_LOG_FILE = $DFLT_LOG ? "#{ENV["HOME"]}/Desktop/lyrics.log" : nil

class Lyrics

	attr_reader :cleanup_lyrics
	attr_writer :cleanup_lyrics
	attr_reader :log_file
	attr_writer :log_file

	def initialize( cleanup_lyrics=true, log_file=$DFLT_LOG_FILE )
		super()
		@cleanup_lyrics = cleanup_lyrics
		@log_file = log_file
	end

	def log?()
		return log_file() != nil
	end

	def long_log?()
		return log?() && $LONG_LOG
	end

	def log( message, new_lines=1 )
		Logging.log( log_file(), message, new_lines )
	end

	def site_host()
		return self.class.site_host()
	end

	def site_name()
		return self.class.site_name()
	end

	def Lyrics.build_song_add_url( artist, title, album=nil, year=nil )
		return nil
	end

	def build_song_add_url( artist, title, album=nil, year=nil )
		return self.class.build_song_add_url?( artist, title, album=nil, year=nil )
	end

	def Lyrics.known_url?( url )
		return url.index( "http://#{site_host}" ) == 0
	end

	def known_url?( url )
		return self.class.known_url?( url )
	end

	def notify( message )
		puts "#{site_name()}: #{message}"
	end

	def Lyrics.normalize_lyrics_data( lyrics_data, artist, title, album, year, url=nil, site_name=nil, site_host=nil )
		lyrics_data["lyrics"] = nil if ( ! lyrics_data.include?( "lyrics" ) || lyrics_data["lyrics"].empty?() )
		lyrics_data["artist"] = artist if ( ! lyrics_data.include?( "artist" ) )
		lyrics_data["title"] = title if ( ! lyrics_data.include?( "title" ) )
		lyrics_data["album"] = album if ( ! lyrics_data.include?( "album" ) )
		lyrics_data["year"] = year if ( ! lyrics_data.include?( "year" ) )
		lyrics_data["url"] = url if ( ! lyrics_data.include?( "url" ) )
		lyrics_data["site_name"] = site_name if ( ! lyrics_data.include?( "site_name" ) )
		lyrics_data["site_host"] = site_host if ( ! lyrics_data.include?( "site_host" ) )
		lyrics_data["suggestions"] = [] if ( ! lyrics_data.include?( "suggestions" ) )
		lyrics_data["custom"] = {} if ( ! lyrics_data.include?( "custom" ) )
		if ( ! lyrics_data.include?( "add_url" ) )
			lyrics_data["add_url"] = build_song_add_url(
				lyrics_data["artist"],
				lyrics_data["title"],
				lyrics_data["album"],
				lyrics_data["year"]
			)
		end

		Strings.decode_htmlentities!( lyrics_data )

		return lyrics_data
	end

	def normalize_lyrics_data( lyrics_data, artist, title, album, year, url=nil )
		self.class.normalize_lyrics_data( lyrics_data, artist, title, album, year, url, site_name(), site_host() )
	end

	# Returns { url, [post] }
	def build_lyrics_fetch_data( artist, title, album=nil, year=nil )
		return { "url" => nil }
	end

	# Returns the lyrics page body or nil if it couldn't be retrieved
	def fetch_lyrics_page( fetch_data )
		return nil, nil if ( fetch_data["url"] == nil )
		log( "Fetching lyrics page... ", 0 ) if ( log?() )
		if ( fetch_data["post"] )
			response, url = HTTP.fetch_page_post( fetch_data["url"], fetch_data["post"], fetch_data["headers"] )
		else
			response, url = HTTP.fetch_page_get( fetch_data["url"], fetch_data["headers"] )
		end
		if ( log?() )
			if ( response && response.body() )
				log( "OK" )
				log( response.body(), 2 ) if ( long_log?() )
			else
				log( "ERROR" )
			end
		end
		return (response ? response.body() : nil), url
	end

	# NOTE: override for example if your using google feeling lucky to verify that
	# the resulting page is not just some random (trashy) result.
	# Returns false is page_body does not correspond to the required lyrics page
	def check_lyrics_page( page_body, artist, title, album=nil, year=nil )
		return true
	end

	# NOTE: implement in subclasses
	# Returns { lyrics, [artist], [title], [album], [year], [custom_data] }
	# def parse_lyrics( url, body, artist, title, album=nil, year=nil )
	# 	return { "lyrics" => nil }
	# end

	# Returns { lyrics, artist, title, album, year, url, site, suggestions, custom_data }
	def lyrics_direct_search( artist, title, album=nil, year=nil )

		fetch_data = build_lyrics_fetch_data( artist, title, album, year )

		if ( log?() )
			log( site_name().upcase() )
			log( "Attempting LYRICS DIRECT SEARCH..." )
			log( " - received artist: #{artist}" )
			log( " - received title: #{title}" )
			log( " - received album: #{album}" ) if ( album )
			log( " - received year: #{year}" ) if ( year )
			log( " - built lyrics url: #{fetch_data["url"]}" )
			if ( fetch_data["post"] && fetch_data["post"].size() > 0 )
				log( " - post data:" )
				post.each { |key, val| log( "    - #{key}: #{val}" ) }
			end
		end

		page_body, page_url = fetch_lyrics_page( fetch_data )

		if ( page_body )
			if ( check_lyrics_page( page_body, artist, title, album, year ) )
				log( "Parsing lyrics... ", 0 ) if ( log?() )
				lyrics_data = parse_lyrics( page_url, page_body, artist, title, album, year )
				if ( log?() )
					log( lyrics_data["lyrics"] ? "LYRICS FOUND" : "LYRICS NOT FOUND" )
					log( " - parsed lyrics:\n[#{lyrics_data["lyrics"]}]" ) if long_log?()
					log( " - parsed artist: #{lyrics_data["artist"]}") if ( lyrics_data["artist"] )
					log( " - parsed title: #{lyrics_data["title"]}") if ( lyrics_data["title"] )
					log( " - parsed album: #{lyrics_data["album"]}") if ( lyrics_data["album"] )
					log( " - parsed year: #{lyrics_data["year"]}") if ( lyrics_data["year"] )
					if ( lyrics_data["custom"] && lyrics_data["custom"].size() > 0 )
						log( " - parsed custom data:" )
						lyrics_data["custom"].each { |key, val| log( "    - #{key}: #{val}" ) }
					end
				end
				lyrics_data["lyrics"] = Strings.cleanup_lyrics( lyrics_data["lyrics"] ) if ( lyrics_data["lyrics"] && @cleanup_lyrics )
				return normalize_lyrics_data( lyrics_data, artist, title, album, year, page_url )
			else
				log( "LYRICS PAGE IS INVALID" ) if ( log?() )
			end
		end

		return normalize_lyrics_data( {}, artist, title, album, year )
	end

	# Returns { lyrics, artist, title, album, year, custom_data }
	def lyrics_from_url( url, artist, title, album=nil, year=nil )

		if ( ! known_url?( url ) )
			if ( log?() )
				log( site_name().upcase() )
				log( "Can't handle received URL: " + url )
			end
			return normalize_lyrics_data( {}, artist, title, album, year )
		end

		if ( log?() )
			log( site_name().upcase() )
			log( "Retrieving LYRICS FROM URL..." )
			log( " - received url: #{url}" )
			log( " - received artist: #{artist}" )
			log( " - received title: #{title}" )
			log( " - received album: #{album}" ) if ( album )
			log( " - received year: #{year}" ) if ( year )
		end

		page_body, page_url = fetch_lyrics_page( { "url" => url } )

		if ( page_body )
			log( "Parsing lyrics... ", 0 ) if ( log?() )
			lyrics_data = parse_lyrics( page_url, page_body, artist, title, album, year )
			if ( log?() )
				log( lyrics_data["lyrics"] ? "LYRICS FOUND" : "LYRICS NOT FOUND" )
				log( " - parsed lyrics:\n[#{lyrics_data["lyrics"]}]" ) if long_log?()
				log( " - parsed artist: #{lyrics_data["artist"]}") if ( lyrics_data["artist"] )
				log( " - parsed title: #{lyrics_data["title"]}") if ( lyrics_data["title"] )
				log( " - parsed album: #{lyrics_data["album"]}") if ( lyrics_data["album"] )
				log( " - parsed year: #{lyrics_data["year"]}") if ( lyrics_data["year"] )
				if ( lyrics_data["custom"] && lyrics_data["custom"].size() > 0 )
					log( " - parsed custom data:" )
					lyrics_data["custom"].each { |key, val| log( "    - #{key}: #{val}" ) }
				end
			end
			lyrics_data["lyrics"] = Strings.cleanup_lyrics( lyrics_data["lyrics"] ) if ( lyrics_data["lyrics"] && @cleanup_lyrics )
			return normalize_lyrics_data( lyrics_data, artist, title, album, year, page_url )
		end

		return normalize_lyrics_data( {}, artist, title, album, year )

	end

	# Returns { url, [post] }
	def build_suggestions_fetch_data( artist, title, album=nil, year=nil )
		return { "url" => nil }
	end

	# Returns the lyrics page body (nil if it couldn't be retrieved) and the response url
	def fetch_suggestions_page( fetch_data )
		return nil, nil if ( fetch_data["url"] == nil )
		if ( log?() )
			log( " - built suggestions url: #{fetch_data["url"]}" )
			if ( fetch_data["post"] )
				log( " - post data:" )
				fetch_data["post"].each { |key, val| log( "    - #{key}: #{val}" ) }
			end
			log( "Fetching suggestions page... ", 0 )
		end
		if ( fetch_data["post"] )
			response, url = HTTP.fetch_page_post( fetch_data["url"], fetch_data["post"], fetch_data["headers"] )
		else
			response, url = HTTP.fetch_page_get( fetch_data["url"], fetch_data["headers"] )
		end
		if ( log?() )
			if ( response && response.body() )
				log( "OK" )
				log( response.body(), 2 ) if ( long_log?() )
			else
				log( "ERROR" )
			end
		end
		return (response ? response.body() : nil), url
	end

	# NOTE: override for example if your using google feeling lucky to verify that
	# the resulting page is not just some random (trashy) result.
	# Returns false is page_body does not correspond to the required suggestions page
	def check_suggestions_page( page_body, artist, title, album=nil, year=nil )
		return true
	end

	# NOTE: implement in subclasses
	# Returns an array of maps { url, artist, title }
	# def parse_suggestions( url, body, artist, title, album=nil, year=nil )
	# 	return []
	# end

	# Returns an array of maps { url, artist, title }
	def suggestions( artist, title, album=nil, year=nil )

		if ( log?() )
			log( site_name().upcase() )
			log( "Retrieving SUGGESTIONS..." )
			log( " - received artist: #{artist}" )
			log( " - received title: #{title}" )
			log( " - received album: #{album}" ) if ( album )
			log( " - received year: #{year}" ) if ( year )
		end

		fetch_data = build_suggestions_fetch_data( artist, title, album, year )
		page_body, page_url = fetch_suggestions_page( fetch_data )
		return [] if ( page_body == nil )

		if ( check_suggestions_page( page_body, artist, title, album, year ) )
			log( "Parsing suggestions...", 0 ) if ( log?() )
			suggestions = parse_suggestions( page_url, page_body, artist, title, album, year )
			suggestions = [] if ( ! suggestions.is_a?( Array ) )
			if ( log?() )
				log( suggestions.empty?() ? " NO SUGGESTIONS FOUND" : "" )
				suggestions.each(){ |sugg| log( " - art: #{sugg["artist"]} | tit: #{sugg["title"]} | url: #{sugg["url"]}" ) }
			end
			return suggestions
		else
			log( "SUGGESTIONS PAGE IS INVALID" ) if ( log?() )
			return []
		end
	end

	# Returns { lyrics, artist, title, album, year, url, site, suggestions, custom_data }
	def on_lyrics_not_found( artist, title, album, year )
		return normalize_lyrics_data( {}, artist, title, album, year )
	end

	def lyrics_from_suggestions( artist, title, album=nil, year=nil, suggestions=nil )

		if ( log?() )
			log( site_name().upcase() )
			log( "Searching LYRICS FROM SUGGESTIONS..." )
		end

		normalized_artist = Strings.normalize( artist )
		normalized_title = Strings.normalize( title )
		suggestions = suggestions( artist, title, album, year ) if ( suggestions == nil )
		log( "Scanning suggestions... ", 0 ) if ( log?() )
		suggestions.each() do |sugg|
			next if ( sugg.class != Hash )
			if ( Strings.normalize( sugg["artist"] ) == normalized_artist &&
				 Strings.normalize( sugg["title"] ) == normalized_title )
				if ( log?() )
					log( "MATCH FOUND" )
					log( " - art: #{sugg["artist"]} | tit: #{sugg["title"]} | url: #{sugg["url"]}" )
				end
				lyrics_data = lyrics_from_url( sugg["url"], sugg["artist"], sugg["title"], album, year )
				if ( lyrics_data["lyrics"] )
					lyrics_data["suggestions"] = suggestions
					return normalize_lyrics_data( lyrics_data, artist, title, album, year, sugg["url"] )
				end
			end
		end
		log( "NO VALID MATCH FOUND" ) if ( log?() )
		return normalize_lyrics_data( { "suggestions" => suggestions }, artist, title, album, year )
	end

	# Returns { lyrics, artist, title, album, year, url, site, suggestions, custom_data }
	def lyrics_full_search( artist, title, album=nil, year=nil )

		# LYRICS DIRECT SEARCH:
		lyrics_data = lyrics_direct_search( artist, title, album, year )
		return lyrics_data if ( lyrics_data["lyrics"] )

		# NOT FOUND, SEARCH IN SUGGESTIONS:
		suggests = suggestions( artist, title, album, year )
		lyrics_data = lyrics_from_suggestions( artist, title, album, year, suggests )
		return lyrics_data if ( lyrics_data["lyrics"] )

		# NOT FOUND, TRY OTHER METHODS
		lyrics_data = on_lyrics_not_found( artist, title, album, year )
		suggests.insert( 0, site_name() ) if ( lyrics_data["suggestions"].size > 0 )
		suggests << "other"
		lyrics_data["suggestions"] = suggests.concat( lyrics_data["suggestions"] )
		return lyrics_data
	end

	# default implementation, overide if there is a more accurate expression for a specific site
	def Lyrics.build_google_feeling_lucky_url( artist, title=nil )
		query = Strings.google_search_quote( artist )
		query << " " << Strings.google_search_quote( title ) if ( title )
		query << " lyrics"
		return Strings.build_google_feeling_lucky_url( query, self.site_host() )
	end

	def build_google_feeling_lucky_url( artist, title=nil )
		return self.class.build_google_feeling_lucky_url( artist, title )
	end

	# NOTE: if possible use popular song that are likely to be indexed by google.
	# We are not testing here how good or complete a site is but if the site plugin
	# is behaving as expected. We part from the assumption that every plugin should
	# be able to get lyrics and suggestions for the values returned here.
	def Lyrics.get_test_data()
		return [
			{ "artist"=>"Nirvana", "album"=>"Nevermind", "title"=>"Smells Like Teen Spirit" },
			{ "artist"=>"Radiohead", "album"=>"Kid A", "title"=>"Optimistic" },
			{ "artist"=>"Massive Attack", "album"=>"Protection", "title"=>"Protection" },
			{ "artist"=>"Pearl Jam", "album"=>"Ten", "title"=>"Porch" },
# 			{ "artist"=>"The Cranberries", "album"=>"Everybody Else Is Doing It, So Why Can't We?", "title"=>"Linger" },
# 			{ "artist"=>"Radiohead", "album"=>"Kid A", "title"=>"Idioteque" },
# 			{ "artist"=>"Radiohead", "album"=>"Kid A", "title"=>"How To Disappear Completely","stitle"=>"How To Dissapear" },
# 			{ "artist"=>"Pearl Jam", "album"=>"Riot Act", "title"=>"Love Boat Captain", "stitle"=>"Love Boat" },
# 			{ "artist"=>"Blur", "album"=>"13", "title"=>"No Distance Left To Run", "stitle"=>"No Distance Left" },
# 			{ "artist"=>"Opeth", "album"=>"Damnation", "title"=>"Hope Leaves" },
# 			{ "artist"=>"Megadeth", "album"=>"Youthanasia", "title"=>"À tout le monde" },
# 			{ "artist"=>"Metallica", "album"=>"Master of Puppets", "title"=>"Master of Puppets" },
# 			{ "artist"=>"Lacuna Coil", "album"=>"Comalies", "title"=>"Heaven's a Lie" },
# 			{ "artist"=>"Pearl Jam", "album"=>"Ten", "title"=>"Porch" },
# 			{ "artist"=>"Portishead", "album"=>"Dummy", "title"=>"Wandering Star" },
		]
	end

	def get_test_data()
		return self.class.get_test_data()
	end

end
